/*

  t-mand.c

  Author: Mika Kojo <mkojo@ssh.fi>

  Copyright (c) 2000 SSH Communications Security Corp
  All rights reserved.

  Created Thu Jan 11 00:22:18 2001.

  */

#include "sshincludes.h"
#include "sshmp.h"

char *outfile = "t-mand.ppm";

void print(char *str, SshMPFloat op)
{
  char *t;
  t = ssh_mprf_get_str(op);
  fprintf(stderr, "%s = %s\n", str, t);
  ssh_xfree(t);
}

/* Simple routine to write a file. */
void write_ppm(char *filename,
               unsigned int *bitmap, size_t xsize, size_t ysize)
{
  FILE *fp;
  int i;

  fp = fopen(filename, "wb");
  if (fp == NULL)
    {
      fprintf(stderr, "write_ppm: could not open file.\n");
      return;
    }

  fprintf(fp, "P3\n# %s\n%u %u %u\n", filename,
          xsize, ysize, 255);
  for (i = 0; i < xsize * ysize; i++)
    {
      unsigned int r,g,b;
      r = bitmap[i] & 0xff;
      g = (bitmap[i] >> 8) & 0xff;
      b = (bitmap[i] >> 16) & 0xff;

      if (i > 0 && (i % 4) == 0)
        {
          fprintf(fp, "\n");
        }
      else
        fprintf(fp, " ");
      fprintf(fp, "%u %u %u", r, g, b);
    }
  fprintf(fp, "\n# %s\n", filename);

  fclose(fp);
}

void qt_pow(unsigned int e, double *z0, double *z1, double *z2, double *z3)
{
  double x0, x1, x2, x3, y0, y1, y2, y3, t0, t1, t2, t3, k;

  x0 = *z0;
  x1 = *z1;
  x2 = *z2;
  x3 = *z3;

  y0 = 1.0;
  y1 = 0.0;
  y2 = 0.0;
  y3 = 0.0;

  while (e)
    {
      if (e&1)
        {
          t0 = y0*x0 - y1*x1 - y2*x2 - y3*x3;
          t1 = y0*x1 + y1*x0 + y2*x3 - y3*x2;
          t2 = y0*x2 + y2*x0 + y3*x1 - y1*x3;
          t3 = y0*x3 + y3*x0 + y1*x2 - y2*x1;

          x0 = t0; x1 = t1; x2 = t2; x3 = t3;
        }

      k   = 2*y0;
      t0  = y0*y0 - y1*y1 - y2*y2 - y3*y3;
      t1 *= k;
      t2 *= k;
      t3 *= k;

      y0 = t0; y1 = t1; y2 = t2; y3 = t3;

      e >>= 1;
    }

  *z0 = x0;
  *z1 = x1;
  *z2 = x2;
  *z3 = x3;
}


void render(unsigned int *ptr, unsigned int xsize, unsigned int ysize)
{
  unsigned int *bitmap;
  unsigned int i, x, y;
  double inv_xsize, inv_ysize;

  bitmap = ssh_xmalloc(sizeof(unsigned int)*xsize*ysize);

  inv_xsize = 1.0/xsize;
  inv_ysize = 1.0/ysize;
  
  for (i = 0, y = 0; y < ysize; y++)
    for (x = 0; x < xsize; x++, i++)
      {
        double r, g, b, t;
        unsigned int u, v, w;

#if 0
        r = ptr[i];
        g = 1.0;
        b = 1.0;
        t = 1.0; 

        qt_pow(2, &r, &g, &b, &t);

        u = r*256.0;
        v = g*256.0;
        w = b*256.0;
#else
        u = ptr[i] & 0x3f;
        v = (ptr[i] >> 6) & 0x3f;
        w = (ptr[i] >> 12) & 0x3f;
        
        u <<= 2;
        v <<= 2;
        w <<= 2;
#endif
        
        bitmap[i] = u | (v << 8) | (w << 16);
      }

  write_ppm(outfile, bitmap, xsize, ysize);
  ssh_xfree(bitmap);
}


void mand(unsigned int xsize, unsigned int ysize,
          SshMPFloat xmin, SshMPFloat ymin,
          SshMPFloat xmax, SshMPFloat ymax,
          unsigned int itr,
          unsigned int *ptr)
{
  SshMPFloatStruct si, sr, ar, ai, fr, fi, br, bi, ax, ay, t;
  unsigned int x, y, i, verb;

  ssh_mprf_init_inherit_prec(&si, xmin);
  ssh_mprf_init_inherit_prec(&sr, xmin);
  ssh_mprf_init_inherit_prec(&ar, xmin);
  ssh_mprf_init_inherit_prec(&ai, xmin);
  ssh_mprf_init_inherit_prec(&fr, xmin);
  ssh_mprf_init_inherit_prec(&fi, xmin);
  ssh_mprf_init_inherit_prec(&br, xmin);
  ssh_mprf_init_inherit_prec(&bi, xmin);
  ssh_mprf_init_inherit_prec(&ax, xmin);
  ssh_mprf_init_inherit_prec(&ay, xmin);
  ssh_mprf_init_inherit_prec(&t,  xmin);

  /* Compute the step size. */
  ssh_mprf_sub(&ax, xmax, xmin);
  ssh_mprf_sub(&ay, ymax, ymin);
  ssh_mprf_set_ui(&t, xsize);
  print("t", &t);
  ssh_mprf_div(&ax, &ax, &t);
  ssh_mprf_set_ui(&t, ysize);
  print("t", &t);
  ssh_mprf_div(&ay, &ay, &t);

  print("xmin", xmin);
  print("ymin", ymin);
  print("xmax", ymax);
  print("ymax", ymax);
  print("ax", &ax);
  print("ay", &ay);
  
  /* Set up the iteration. */
  ssh_mprf_set(&si, ymin);

  for (y = 0; y < ysize; y++)
    {
      printf("rows left: %u\r", ysize - y);
      fflush(stdout);
      ssh_mprf_set(&sr, xmin);
      for (x = 0; x < xsize; x++)
        {
          ssh_mprf_square(&ar, &sr);
          ssh_mprf_square(&ai, &si);
          ssh_mprf_set(&fr, &sr);
          ssh_mprf_set(&fi, &si);

          for (i = 0; i < itr; i++)
            {
              ssh_mprf_add(&t, &ar, &ai);
              if (ssh_mprf_cmp_ui(&t, 4) >= 0)
                break;

              /* br = ar - ai + sr */
              ssh_mprf_sub(&br, &ar, &ai);
              ssh_mprf_add(&br, &br, &sr);

              /* bi = fr*fi*2 + si */
              ssh_mprf_mul(&bi, &fr, &fi);
              ssh_mprf_mul_2exp(&bi, &bi, 1);
              ssh_mprf_add(&bi, &bi, &si);

              ssh_mprf_square(&ar, &br);
              ssh_mprf_square(&ai, &bi);
              ssh_mprf_set(&fr, &br);
              ssh_mprf_set(&fi, &bi);
            }

          ptr[x + y*xsize] = itr - i;
          ssh_mprf_add(&sr, &sr, &ax);
        }
      ssh_mprf_add(&si, &si, &ay);
    }
  
  ssh_mprf_clear(&si);
  ssh_mprf_clear(&sr);
  ssh_mprf_clear(&ar);
  ssh_mprf_clear(&ai);
  ssh_mprf_clear(&fr);
  ssh_mprf_clear(&fi);
  ssh_mprf_clear(&br);
  ssh_mprf_clear(&bi);
  ssh_mprf_clear(&ax);
  ssh_mprf_clear(&ay);
  ssh_mprf_clear(&t);

  return;
}


void makepic(unsigned int xsize, unsigned int ysize,
             unsigned int iter)
{
  SshMPFloatStruct xmin, xmax, ymin, ymax;
  unsigned int *ptr;

  ptr = ssh_xmalloc(xsize * ysize * sizeof(unsigned int));

  ssh_mprf_init_with_prec(&xmin, 10);
  ssh_mprf_init_with_prec(&xmax, 10);
  ssh_mprf_init_with_prec(&ymin, 10);
  ssh_mprf_init_with_prec(&ymax, 10);

  ssh_mprf_set_si(&xmin, -2);
  ssh_mprf_set_si(&xmax,  1);
  ssh_mprf_set_si(&ymin, -2);
  ssh_mprf_set_si(&ymax,  1);

  mand(xsize, ysize, &xmin, &ymin, &xmax, &ymax, iter, ptr);

  render(ptr, xsize, ysize);

  ssh_xfree(ptr);
  
  ssh_mprf_clear(&xmin);
  ssh_mprf_clear(&xmax);
  ssh_mprf_clear(&ymin);
  ssh_mprf_clear(&ymax);
}

int main(int ac, char *av[])
{
  unsigned int xsize, ysize, iter, i;

  /* Defaults. */
  xsize = 100;
  ysize = 100;
  iter  = 32;
  
  for (i = 1; i < ac; i++)
    {
      if (strcmp(av[i], "-x") == 0)
        {
          xsize = atoi(av[i+1]);
          i++;
          continue;
        }
      if (strcmp(av[i], "-y") == 0)
        {
          ysize = atoi(av[i+1]);
          i++;
          continue;
        }
      if (strcmp(av[i], "-i") == 0)
        {
          iter = atoi(av[i+1]);
          i++;
          continue;
        }
      fprintf(stderr, "error: unknown token '%s'.\n", av[i]);
      exit(1);
    }

  fprintf(stderr, "Making a picture...\n");
  makepic(xsize, ysize, iter);
  return 0;
}

/* t-mand.c */






















